#include "stc.h"
#include "usb.h"
#include "usb_req_std.h"
#include "usb_req_class.h"
#include "usb_req_vendor.h"
#include "util.h"
#include "uart.h"

BYTE DeviceState;
SETUP Setup;
EPSTATE Ep0State;
BYTE InEpState;
BYTE OutEpState;

void usb_init()
{
    USBCLK = 0x00;		// λֵ0x20,0010 0000,PLLʱѡ24MPLLʱѡ12M
						// 32G0x20ָPLLʱԴѡ1/2,ʱԴѡ 1/1.ûPLLƺû
    USBCON = 0x90;		// λֵ=0.ENUSB	USBRST	PS2M	PUEN	PDEN	DFREC	DP	DM
						// ֵ0x90 ENUSB = 1;USBж/ʱPUEN =1;ڲ1.5k
    
    usb_write_reg(FADDR, 0x00);
    usb_write_reg(POWER, 0x08);			// USBRST=1;USBλ
    usb_write_reg(INTRIN1E, 0x3f);		// IN?ϵжϣ3Fȫ
    usb_write_reg(INTROUT1E, 0x3f);		// OUT?ϵжϣ3FȫˡEP0_OUTûдһ˵
    usb_write_reg(INTRUSBE, 0x07);		// USBԴжĴ - - - - SOFIE RSTIE RSUIE SUSIE
										// ֱǣ֡ʼźж/λж/ָźж/źж
    usb_write_reg(POWER, 0x00);

    DeviceState = DEVSTATE_DEFAULT;		// 豸Ĭ״̬
    Ep0State.bState = EPSTATE_IDLE;		// ϵ״̬
    InEpState = 0x00;					// In˵״̬
    OutEpState = 0x00;					// Out˵״̬

    IE2 |= EUSB;
}

BYTE usb_read_reg(BYTE addr)
{
	BYTE dat;

	while (USBADR & 0x80);
	USBADR = addr | 0x80;
	while (USBADR & 0x80);
	dat = USBDAT;

	return dat;
}

void usb_write_reg(BYTE addr, BYTE dat)
{
	while (USBADR & 0x80);
	USBADR = addr & 0x7f;
	USBDAT = dat;
}

BYTE usb_read_fifo(BYTE fifo, BYTE *pdat)
{
    BYTE cnt;
    BYTE ret;
	ret = cnt = usb_read_reg(COUNT0);
	
    while (cnt--)
    {
    	*pdat++ = usb_read_reg(fifo);
    }

    return ret;
}

void usb_write_fifo(BYTE fifo, BYTE *pdat, BYTE cnt) reentrant
{
	while (cnt--)
	{
        usb_write_reg(fifo, *pdat++);
    }
}

void usb_isr() interrupt 25
{
    BYTE intrusb;
    BYTE intrin;
    BYTE introut;

    intrusb = usb_read_reg(INTRUSB);
    intrin = usb_read_reg(INTRIN1);
    introut = usb_read_reg(INTROUT1);

    if (intrusb & RSUIF) usb_resume();
    if (intrusb & RSTIF) usb_reset();
    if (intrin & EP0IF) usb_setup();

    if (intrin & 0x3e) usb_in_epn();
    if (introut & 0x3e) usb_out_epn();

    if (intrusb & SUSIF) usb_suspend();
}

void usb_resume()
{
}

void usb_reset()
{
    usb_write_reg(FADDR, 0x00);
    DeviceState = DEVSTATE_DEFAULT;
    Ep0State.bState = EPSTATE_IDLE;

    usb_write_reg(INDEX, epin);
    usb_write_reg(INCSR1, INCLRDT | INFLUSH);
    usb_write_reg(INDEX, epout);
    usb_write_reg(OUTCSR1, OUTCLRDT | OUTFLUSH);

    usb_write_reg(INDEX, 0);
}

void usb_suspend()
{
}

void usb_setup()
{
    BYTE csr;

    usb_write_reg(INDEX, 0);
    csr = usb_read_reg(CSR0);

    if (csr & STSTL)
    {
        usb_write_reg(CSR0, csr & ~SDSTL);
        Ep0State.bState = EPSTATE_IDLE;
    }
    if (csr & SUEND)
    {
        usb_write_reg(CSR0, csr & ~SSUEND);
    }

    switch (Ep0State.bState)
    {
    case EPSTATE_IDLE:
        if (csr & OPRDY)
        {
            usb_read_fifo(FIFO0, (BYTE *)&Setup);
            Setup.wLength = reverse2(Setup.wLength);
            switch (Setup.bmRequestType & REQUEST_MASK)
            {
            case STANDARD_REQUEST:
                usb_req_std();
                break;
            case CLASS_REQUEST:
                usb_req_class();
                break;
            case VENDOR_REQUEST:
                usb_req_vendor();
                break;
            default:
                usb_setup_stall();
                return;
            }
        }
        break;
    case EPSTATE_DATAIN:
        usb_ctrl_in();
        break;
    case EPSTATE_DATAOUT:
        usb_ctrl_out();
        break;
    }
}

void usb_setup_stall()
{
    Ep0State.bState = EPSTATE_STALL;
    usb_write_reg(CSR0, SOPRDY | SDSTL);
}

void usb_setup_in()
{
    Ep0State.bState = EPSTATE_DATAIN;
    usb_write_reg(CSR0, SOPRDY);
    usb_ctrl_in();
}

void usb_setup_out()
{
    Ep0State.bState = EPSTATE_DATAOUT;
    usb_write_reg(CSR0, SOPRDY);
}

void usb_setup_status()
{
    Ep0State.bState = EPSTATE_IDLE;
    usb_write_reg(CSR0, SOPRDY | DATEND);
}

void usb_ctrl_in()
{
    BYTE csr;
    BYTE cnt;

    usb_write_reg(INDEX, 0);
    csr = usb_read_reg(CSR0);
    if (csr & IPRDY) return;

    cnt = Ep0State.wSize > EP0_SIZE ? EP0_SIZE : Ep0State.wSize;
    usb_write_fifo(FIFO0, Ep0State.pData, cnt);
    Ep0State.wSize -= cnt;
    Ep0State.pData += cnt;
    if (Ep0State.wSize == 0)
    {
        usb_write_reg(CSR0, IPRDY | DATEND);
        Ep0State.bState = EPSTATE_IDLE;
    }
    else
    {
        usb_write_reg(CSR0, IPRDY);
    }
}

void usb_ctrl_out()
{
    BYTE csr;
    BYTE cnt;

    usb_write_reg(INDEX, 0);
    csr = usb_read_reg(CSR0);
    if (!(csr & OPRDY)) return;

    cnt = usb_read_fifo(FIFO0, Ep0State.pData);
    Ep0State.wSize -= cnt;
    Ep0State.pData += cnt;
    if (Ep0State.wSize == 0)
    {
        usb_write_reg(CSR0, SOPRDY | DATEND);
        Ep0State.bState = EPSTATE_IDLE;
    }
    else
    {
        usb_write_reg(CSR0, SOPRDY);
    }

    usb_uart_settings();
}

void usb_bulk_intr_in(BYTE *pData, BYTE bSize, BYTE ep)
{
    usb_write_fifo((BYTE)(FIFO0 + ep), pData, bSize);
    usb_write_reg(INCSR1, INIPRDY);
}

void usb_in_epn()
{
    BYTE csr;

    usb_write_reg(INDEX, epin);
    csr = usb_read_reg(INCSR1);
    if (csr & INSTSTL)
    {
        usb_write_reg(INCSR1, INCLRDT);
    }
    if (csr & INUNDRUN)
    {
        usb_write_reg(INCSR1, 0);
    }
}

void usb_out_epn()
{
    BYTE csr;

    usb_write_reg(INDEX, epout);
    csr = usb_read_reg(OUTCSR1);
    if (csr & OUTSTSTL)								// STSTL(sent stall)stallźŷ
    {
        usb_write_reg(OUTCSR1, OUTCLRDT);			// CLRDTλ1STALL־
    }
    if (csr & OUTOPRDY)								// OPRDYݰɱ־жݺӦýλ->uart_recv_done()ȥ
    {
        RxCount = usb_read_fifo((BYTE)(FIFO0 + epout), RxBuffer);	// յȫ뻺RxBufferֽRxCount
        RxFlag = 1;									// Ѿݽϵȴ־
    }
}
